# -*- encoding: utf-8 -*-
#

"""
This is Wombat, an operations and maintenance tool for Erlang systems
© 2015, Erlang Solutions Ltd
"""

from __future__ import absolute_import

import cmd
import sys
from pprint import pprint
from wombat.arg_parse import ArgParse
from wombat.wombat import WombatException, Wombat
from wombat.utils import then_and_now, timestr

def ppprint(d):
    """Pretty print a dict object as `key value' lines, given a list recurse"""
    if type(d) == list:
        for i in d:
            ppprint(i)
            print()
    elif type(d) == dict:
        l = max([len(str(k)) for k in d]) + 2
        fmt = '{:<' + str(l) + '}{!s}'
        for k in sorted(d):
            print(fmt.format(k, d[k]))
    else:
        pprint(d)

def flatten_metric_info(metric_info_groups):
    return [metric for group in metric_info_groups
            for cat in group['categories']
            for metric in cat['metrics']]

class WombatCLI(cmd.Cmd):
    """CLI interface for accessing wombat functionality"""

    def __init__(self, w):
        """Initialize with a Wombat() instance"""
        cmd.Cmd.__init__(self)
        self.wombat = w
        self.prompt = 'Wombat@' + w.host + '> '
        self.intro = w.header

    def get(self, path, format = 'python'):
        """Do an HTTP GET towards Wombat,
        and display an error message on exception"""
        try:
            return self.wombat.get(path, format)
        except WombatException as we:
            print(str(we))
            return None

    def post(self, path, data, format = 'python'):
        """Do an HTTP POST towards Wombat,
        and display an error message on exception"""
        try:
            return self.wombat.post(path, data, format)
        except WombatException as we:
            print(str(we))
            return None

    def emptyline(self):
        """cmd.Cmd override, ignore empty lines"""
        return False

    def default(self, line):
        """cmd.Cmd override, display help message"""
        print("Type `help' for a list of command or `exit' to quit")

    def do_get(self, arg):
        """get PATH
        Get data from Wombat"""
        data = self.get(arg)
        if data:
            pprint(data)

    def do_rawget(self, arg):
        """rawget PATH
        Get data from Wombat as JSON"""
        data = self.get(arg, format = 'json')
        if data:
            print(data)

    def do_add(self, arg):
        """add node NODE COOKIE"""
        a = arg.strip()

        args = a.split()
        if args[0] == 'node' and len(args) == 3:
            r = self.post('topo/action/add-node',
                           {'name': args[1],
                             'cookie': args[2]})
        else:
            print('Wrong arguments')
            return

        if not r: return
        ppprint(r)
        print()

    def do_delete(self, arg):
        """delete node NODE-ID"""
        args = arg.strip().split()
        if args[0] == 'node' and len(args) == 2:
            r = self.wombat.delete('topo/node/{}'.format(args[1]))
            if r == 204:
                print('204 No Content: DELETE OK')
            else:
                print('ERROR: HTTP response {}'.format(r))
        else:
            print('Wrong arguments')

    def do_list(self, arg):
        """list [ nodes | alarms | families | node-metrics NODE-ID] """
        a = arg.strip()

        if a == 'nodes':
            r = self.get('topo/node')
        elif a == 'families':
            r = self.get('topo/node-family')
        elif a == 'alarms':
            r = self.get('alarm')['alarms']
        else:
            args = a.split()
            if args[0] == 'node-metrics' and len(args) == 2:
                r0 = self.get('metric/metric-info/node/{}'
                                     .format(args[1]))['groups']
                if not r0: return
                r = flatten_metric_info(r0)
            else:
                print('Not enough arguments')
                return

        if not r: return
        ppprint(r)

    def do_exit(self, arg):
        """Leave the CLI"""
        return True

    def do_metric(self, arg):
        """metric NODE-ID METRIC [ TYPE ] [ MINUTES ]

        NODE-ID cannot contain whitespace.

        METRIC is an Erlang term, e.g. `binary_memory` or
        `'{folsom, mymetric}'`, which means that it can contain spaces.

        TYPE is any of the following: counter, gauge, meter, spiral, histogram.

        MINUTES is an integer.
        """
        args = arg.split()
        metric_types = ['counter', 'gauge', 'meter', 'spiral', 'histogram']
        metric_type = metric_types[0]
        minute = 1
        if len(args) < 2:
            print('Not enough arguments')
            return
        else:
            if len(args) > 2 and str.isdigit(args[-1]):
                minute = args.pop()
            if len(args) > 2 and args[-1] in metric_types:
                metric_type = args.pop()
            node = args[0]

            # The METRIC parameter can contain spaces, so we need to join them.
            # FIXME This doesn't work when a METRIC has several spaces next to
            # each other, e.g. `'{folsom, <<"My   metric">>}'`, because then
            # the metric variable wil contain `'{folsom, <<"My metric">>}'`.
            metric = ' '.join(args[1:])
        (fro, to) = then_and_now(int(minute))
        try:
            ms = self.wombat.get_node_metric(node, metric_type,
                                             metric, to, fro)
            if metric_type in ['counter', 'gauge', 'meter', 'spiral']:
                for m in ms:
                    print('{0[time]:20} {0[value]}'.format(m))
            elif metric_type == 'histogram':
                pprint(ms)
        except:
            pass

    def do_cacti_metrics(self, arg):
        """cacti_metrics NODE-ID [METRIC1 ... ]
        Get a list of node metrics for Cacti to collect"""
        a = arg.split()
        node = a[0]
        (fro, to) = then_and_now(15)
        metrics = a[1:]
        if len(metrics) == 1:
            try:
                x = self.wombat.get(
                            'metric/history/node/{}/{}/1?to={}&from={}'
                            .format(node, metrics[0], to, fro))['data'][-1]['value']
                sys.stdout.write(str(x))
            except:
                sys.stdout.write('0')
        else:
            for m in metrics:
                try:
                    x = self.wombat.get(
                               'metric/history/node/{}/{}/1?to={}&from={}'
                               .format(node, m, to, fro))['data'][-1]['value']
                    sys.stdout.write('{}:{} '.format(m, x))
                except:
                    sys.stdout.write('{}:{} '.format(m, 0))

    def do_timequery(self, arg):
        """timequery MINUTES
        generate query string for REST queries of metrics"""
        (fro, to) = then_and_now(int(arg))
        print('?from={}&to={}'.format(fro, to))

def main():
    try:
        import readline
    except:
        pass
    a, rest = ArgParse().parse_known_args()
    w = Wombat(a.host, a.port, a.user, a.password, https=a.https)
    cli = WombatCLI(w)
    if not rest:
        cli.cmdloop()
    else:
        cli.onecmd(' '.join(rest))

# Main program
if __name__ == '__main__':
    main()
