#!/usr/bin/env python3

# -*- encoding: utf-8 -*-
#

"""
This script reads data from Wombat and pushes it into Zabbix.

© 2015-2020, Erlang Solutions Ltd
"""
import pprint
import time
import base64
import sys
import os
import traceback
import subprocess
from subprocess import Popen, PIPE, DEVNULL
from io import StringIO
from datetime import datetime
from wombat.wombat import Wombat
from wombat.arg_parse import ArgParse
from wombat.wombatcli import flatten_metric_info
from wombat.utils import then_and_now
from wombat.wombat_rest import quote_path
from zabbix_api import ZabbixAPI, Already_Exists


class ZabbixException(Exception):
    pass


COLORS = ('00CC00', 'CC0000', '0000CC', 'CC00CC', '00CCCC', 'CCCC00', 'C0C0C0')
ALREADY_EXISTS = 'ZabbixPlugin_Already_Exists'
EXTERNALSCRIPT = 'wombat_getter'


def get_zabbix_color(index):
    """Return a color code for Zabbix graph"""

    return COLORS[index % len(COLORS)]


def do_zabbix(zabbix, request, params):
    """Shortener function for zabbix request"""

    try:
        return zabbix.do_request(
                zabbix.json_obj(request, params=params))['result']
    except Already_Exists:
        return ALREADY_EXISTS


def decode_from_zabbix_key(zabbix_key):
    """Return the string that was encoded to base64 for zabbix items key"""

    b64 = str.encode(zabbix_key.ljust(-(-len(zabbix_key) // 4) * 4, '='))
    return base64.urlsafe_b64decode(b64).decode()


def encode_to_zabbix_key(name, subname=None):
    """Return a zabbix item key.

    It is a base64 code without trailing equal signs.
    """

    name += '.'
    if subname is not None:
        name += subname
    return base64.urlsafe_b64encode(str.encode(name)).decode().rstrip('=')


def get_zabbix_keys(metric, metric_type):
    """Return a dictionary with Wombat names as keys and zabbix keys as values
    """

    def encode(sub):
        return encode_to_zabbix_key(metric, sub)
    if metric_type in ('counter', 'gauge'):
        return {'value': encode('')}
    elif metric_type == 'spiral':
        return {'count': encode('count'), 'one': encode('one')}
    elif metric_type == 'meter':
        return {'day': encode('day'), 'fifteen': encode('fifteen'),
                'five': encode('five'), 'one': encode('one')}


def zabbix_time(wombat_time):
    """Return Wombat's time in zabbix time representation"""

    return int(wombat_time) // 1000


def data_to_sender(host, item_key, time, value):
    """Return a string representation of zabbix data.

    This will be sent with zabbix_sender.
    """

    if isinstance(value, str):
        value = '"' + str(value).replace('\\', '\\\\').replace(
                        '\n', '\\n').replace('"', '\\"') + '"'
    return '"{0}" {1} {2} {3}\n'.format(host, item_key, time, value)


def create_zabbix_data(host, metric_type, zabbix_key, metric_data):
    """Convert metric values from wombat to a string for sending it."""

    datastr = ''
    for m in metric_data:
        time = zabbix_time(m['jstime'])
        for dname in sorted(m['value']):
            datastr += data_to_sender(host, zabbix_key[dname],
                                      time, m['value'][dname])
    return datastr


def get_application_ids(zabbix, apps, hostid, metric):
    """Return an application id list with one element.

    In zabbix each trapper item of Wombat data belongs to an application,
    which is identified by this applicaton id list.
    """

    if apps:
        return apps
    found = do_zabbix(zabbix, 'application.get',
                      {'hostids': hostid, 'filter': {'name': metric}})
    if found and 'applicationid' in found[0]:
        return [found[0]['applicationid']]
    elif found:
        raise ZabbixException('Incorrect format of returned JSON: {!s}'.format(
            found))

    created = do_zabbix(zabbix, 'application.create',
                        {'name': metric, 'hostid': hostid})
    if created and 'applicationids' in created:
        return created['applicationids']
    elif created == ALREADY_EXISTS:
        return get_application_ids(zabbix, apps, hostid, metric)
    else:
        raise ZabbixException('Application creation is failed. '
                              'Returned JSON: {!s}'.format(created))


def add_to_graph(zabbix, hostid, graph_name, itemid, graphid):
    """Return a graph id.

    In zabbix each trapper item of Wombat metric belongs to a graph,
    which is identified by this graph id.
    """

    gitem = {'itemid': itemid, 'color': get_zabbix_color(0), 'sortorder': '0'}
    if not graphid:
        graphs = do_zabbix(zabbix, 'graph.get',
                           {'hostids': hostid, 'filter': {'name': graph_name}})
        if graphs and 'graphid' in graphs[0]:
            graphid = graphs[0]['graphid']
    if graphid:
        graphitems = do_zabbix(zabbix, 'graphitem.get', {'graphids': graphid,
                                                         'output': 'extend'})
        if graphitems:
            gitem['sortorder'] = str(len(graphitems))
            gitem['color'] = get_zabbix_color(len(graphitems))
        gitems = graphitems + [gitem]
        do_zabbix(zabbix, 'graph.update',
                  {'graphid': graphid, 'gitems': gitems})
        return graphid
    else:
        params = {'name': graph_name, 'width': '900',
                  'height': '200', 'gitems': [gitem]}

        created = do_zabbix(zabbix, 'graph.create', params)
        if created and 'graphids' in created and created['graphids']:
            return created['graphids'][0]
        elif created == ALREADY_EXISTS:
            graphs = do_zabbix(zabbix, 'graph.get', {
                           'hostids': hostid, 'filter': {'name': graph_name}})
            return graphs[0]['graphid']
        else:
            raise ZabbixException('Graph creation is failed. '
                                  'Returned JSON: {!s}'.format(created))


def get_items_list(zabbix, hostid):
    """Return available zabbix items' key."""

    return [i['key_'] for i in do_zabbix(zabbix, 'item.get', {
                                    'hostids': hostid, 'output': 'extend'})]


def create_item(zabbix, name, key, hostid, apps=[], item_type=2, value_type=0):
    """Return the newly created item's id."""

    params = {
        'name': name,
        'key_': key,
        'hostid': hostid,
        'type': item_type,
        'value_type': value_type,
        'applications': apps
    },
    created = do_zabbix(zabbix, 'item.create', params)
    

    if created and 'itemids' in created and created['itemids']:
        return created['itemids'][0]
    elif created == ALREADY_EXISTS:
        return ALREADY_EXISTS
    else:
        raise ZabbixException('Item creation is failed. '
                              'Returned JSON: {!s}'.format(created))

def create_trigger(zabbix, alarm_key, encoded_key, hostid, alarmproperties):
    severitycases = {'minor': 4, 'major': 5, 'warning': 2, 'indeterminate': 0}
    hostparam = {'output': ['host'], 'hostids': str(hostid)}
    hostjson = do_zabbix(zabbix, 'host.get', hostparam)
    host = hostjson[0]['host']
    print('Creating trigger')
    params = {
                 'description': alarm_key,
                 'expression': '{' + host + ':' + encoded_key + '.last()}>0',
                 'priority': severitycases[alarmproperties[0]],
                 'comments': alarmproperties[1]
                }

    created = do_zabbix(zabbix, 'trigger.create', params)

    if created and 'triggerids' in created and created['triggerids']:
        return created['triggerids'][0]
    elif created == ALREADY_EXISTS:
        return ALREADY_EXISTS
    else:
        raise ZabbixException('Trigger creation is failed. '
                              'Returned JSON: {!s}'.format(created))

def create_missing_metric_items(zabbix, hostid, trapper_items, metrics):
    """Create a zabbix item for each missing metrics
    with application and graph id."""

    for mtype, metric in metrics:
        zabbix_keys = get_zabbix_keys(metric, mtype)
        apps = []
        graphid = 0
        for _, key in sorted(zabbix_keys.items()):
            if key not in trapper_items:
                apps = get_application_ids(zabbix, apps, hostid, metric)
                name = decode_from_zabbix_key(key).rstrip('.')
                itemid = create_item(zabbix, name, key, hostid, apps)
                if itemid == ALREADY_EXISTS:
                    continue
                graphid = add_to_graph(zabbix, hostid, metric, itemid, graphid)


def create_missing_log_items(zabbix, hostid, trapper_items, loggers):
    """Create a zabbix item for each missing loggers (notifications, alarms)"""

    for logger in loggers:
        key = encode_to_zabbix_key(logger)
        if key not in trapper_items:
            appname = logger.rsplit('.', 1)[0]
            apps = get_application_ids(zabbix, [], hostid, appname)
            create_item(zabbix, logger, key, hostid, apps, value_type=4)

def create_missing_alarm_items(zabbix, hostid, trapper_items, alarms):
    """Create a zabbix item for each missing alarms (notifications, alarms)"""
    for alarm_key in alarms:
        encoded_key = encode_to_zabbix_key(alarm_key)
        if encoded_key not in trapper_items:
            appname = alarm_key.rsplit('.', 1)[0]
            apps = get_application_ids(zabbix, [], hostid, appname)
            create_item(zabbix, alarm_key, encoded_key, hostid, apps)
            create_trigger(zabbix, alarm_key, encoded_key, hostid, alarms[alarm_key])

def clear_alarms(zabbix, hostid, trapper_items, alarms):

    alarms0 = []
    for i in alarms:
        alarms0.append(encode_to_zabbix_key(i))
    alarms = alarms0
    alarms_str = ''
    millis = int(round(time.time() * 1000))
    for trapper in trapper_items:
        if trapper not in alarms:
            atime = zabbix_time(millis)
            alarms_str += data_to_sender(hostid, trapper, atime, 0)
    return alarms_str

def get_metric_info(wombat, node):
    """Return available metrics for a node excluding histograms"""

    deep = wombat.get('metric/metric-info/node/{}'
                      .format(quote_path(node)))['groups']
    if not deep:
        return []
    flat = flatten_metric_info(deep)
    return [(m['type'], m['name']) for m in flat if m['type'] != 'histogram']


def get_wombat_metrics(wombat, node, metrics, time_diff=1):
    """Return zabbix data for sending.

    Queries Wombat for metric series, converts them into strings
    and concatenates them.
    """

    (fro, to) = then_and_now(time_diff)
    datastr = ''
    for mtype, metric in metrics:
        ms = wombat.get_node_metric(node, mtype, metric, to, fro)
        if mtype in ('counter', 'gauge'):
            [m.update({'value': {'value': m['value']}}) for m in ms]
        zabbix_key = get_zabbix_keys(metric, mtype)
        datastr += create_zabbix_data(node, mtype, zabbix_key, ms)
    return datastr


def get_logs(wombat, node, time, time_diff=1):
    """Return zabbix data for sending.

    Queries Wombat for notifications, converts them into strings
    and concatenates them.
    """

    # FIXME: Each minute, we query the notifications of the last minute. So if
    # the timing is off a little, it can happen that some notifications are
    # lost or duplicated.

    cursor = ''
    to = int(time)
    fro = to - time_diff * 60
    currtime = to
    logs = ''
    loggers = []
    while currtime >= fro:
        wo_log_data = wombat.get('logs/node/' + node + cursor)
        for l in wo_log_data['logs']:
            ltime = zabbix_time(l['date'])
            if fro <= ltime < to:
                logger = l['originator'] + '.' + l['severity']
                logs = data_to_sender(node, encode_to_zabbix_key(logger),
                                      ltime, l['message']) + logs
                loggers.append(logger)
            currtime = min(currtime, ltime)
        if not wo_log_data['cursor']:
            break
        cursor = '?c=' + wo_log_data['cursor']
    return list(set(loggers)), logs

def get_alarms(wombat, node, time, time_diff=2):
    alarms_str = ''
    alarms_data = {}
    wo_alarms_data = wombat.get('alarm/node/' + node)
    for a in wo_alarms_data['alarms']:
        atime = zabbix_time(a['eventTime'])
        alarm_key = a['alarmId'] + '.alarm'
        zabbix_key = encode_to_zabbix_key(alarm_key)
        alarms_data.update({alarm_key: []})
        alarms_str += data_to_sender(node, zabbix_key, atime, 1)
        alarms_data[alarm_key].append(a['severity'])
        alarms_data[alarm_key].append(a['additionalInformation'])
    return alarms_data, alarms_str

def parse_with_zabbix_args():
    """Return parsed args, after added some new arguments."""

    parser = ArgParse()
    parser.add_argument('--zabbix_server')
    parser.add_argument('--zabbix_user')
    parser.add_argument('--zabbix_password')
    parser.add_argument('--zabbix_group')
    a, rest = parser.parse_known_args()
    return a


def main():
    """Put data from Wombat to Zabbix"""

    if sys.argv[0].endswith(EXTERNALSCRIPT):
        print('started: {}'.format(time.strftime("%a, %d %b %Y %X",
                                                 time.localtime())))
        print('{}'.format(str(sys.argv[0])))
        print('{}'.format(os.path.realpath(sys.argv[0])))
        plugin_args = [os.path.realpath(sys.argv[0])] + sys.argv[1:] + [
            '&', 'disown']
        Popen(plugin_args, stdin=None, stdout=DEVNULL, stderr=DEVNULL)
        print('stopped: {}'.format(time.strftime("%a, %d %b %Y %X",
                                                 time.localtime())))
        print(sys.argv[0], plugin_args)
        quit()
    start_time = time.time()
    sys.stdout = plugin_logs = StringIO()
    sys.stderr = plugin_logs
    print('args: {}'.format(' '.join(sys.argv)))
    print('python started: {}'.format(time.strftime("%a, %d %b %Y %X",
                                                    time.localtime())))
    try:
        a = parse_with_zabbix_args()
        # FIXME Check if login succeded
        wombat = Wombat(a.host, a.port, a.user, a.password, https=a.https)
        zabbix = ZabbixAPI(server=a.zabbix_server)
        zabbix.login(a.zabbix_user, a.zabbix_password)

        # Get hosts in the zabbix_group group.
        pars = {'selectHosts': ['host'], 'filter': {'name': [a.zabbix_group]}}
        hosts = do_zabbix(zabbix, 'hostgroup.get', pars)[0]['hosts']

        nodes = [(h['host'], h['hostid']) for h in hosts
                 if h['host'] not in (a.zabbix_group, 'wombat')]
        gen_h = [h for h in hosts if h['host'] in (a.zabbix_group, 'wombat')]

        wombat_data = []
        for host, hostid in nodes:

            # Fix me: our sessions auth process is overwelmed with these calls, leaving just alarms here.
            # metrics = get_metric_info(wombat, host)
            # loggers, log_data = get_logs(wombat, host, start_time)
            # metric_data = get_wombat_metrics(wombat, host, metrics)
            alarms, alarm_data = get_alarms(wombat, host, start_time)
            wombat_data += [(hostid, alarm_data, alarms)]
            # FIXME Wait here if another plugin process is running.
        for (hostid, alarm_data, alarms) in wombat_data:
            items = get_items_list(zabbix, hostid)
            # create_missing_metric_items(zabbix, hostid, items, metrics)
            # create_missing_log_items(zabbix, hostid, items, loggers)
            
            create_missing_alarm_items(zabbix, hostid, items, alarms)
            alarm_data += clear_alarms(zabbix, host, items, alarms)

            datastr = '{0}'.format(alarm_data).strip()

            sender_args = ['zabbix_sender', '-z', '127.0.0.1', '-p',
                           '10051', '-T', '-i', '-']
            sender = Popen(sender_args, stdin=PIPE, stdout=PIPE, stderr=PIPE)
            output, error = sender.communicate(str.encode(datastr))

            if output:
                print(output.decode().splitlines(True)[-1].rstrip('\n'))
            if error:
                print(error.decode())
    except:
        print('{!s}'.format(traceback.format_exc()))
    print('python stopped: {}'.format(time.strftime("%a, %d %b %Y %X",
                                                    time.localtime())))
    if gen_host:
        subprocess.call(['zabbix_sender', '-z', '127.0.0.1', '-p', '10051',
                         '-s', gen_host[0]['host'], '-k', 'wo_getter_log',
                         '-o', plugin_logs.getvalue()])

# Main program
if __name__ == '__main__':
    main()
