<!-- -*- markdown -*- -->

# Continuous Integration in Wombat

This page describes how continuous integration was set up for the
Wombat project, what conventions shall developers follow when writing
tests and hopefully provides enough information to migrate CI to a
different Jenkins server if it would become necessary.

## The master and the slave

The Wombat project uses two machines for CI:

* The Jenkins master server (http://jenkins.erlang-solutions.com/). We have two
  accounts on ci-03: `wombat_dev` for developers (to view and run jobs) and
  `wombat_admin` for admins (to configure jobs). We don't have either Jenkins
  admin access or access to the Jenkins machine.

* The Wombat jobs are executed on the ci-03 slave machine. We have ssh and sudo
  access towards ci-03 machine.

Whenever we talk below about the Jenkins web interface, we refer to the Jenkins
master. Whenever we talk about issuing commands on the Jenkins machine, we mean
the ci-03 slave machine.

### Global Jenkins configuration

Jenkins came with a large number of plugins preinstalled. We use the
following ones:

- One or more of the following git plugins (I'm not sure which one is
  providing the necessary functionality):
  - [Git Client Plugin](http://wiki.jenkins-ci.org/display/JENKINS/Git+Client+Plugin)
  - [Git Plugin](http://wiki.jenkins-ci.org/display/JENKINS/Git+Plugin)
  - [GitHub API Plugin](https://wiki.jenkins-ci.org/display/JENKINS/GitHub+API+Plugin)
  - [GitHub plugin](http://wiki.jenkins-ci.org/display/JENKINS/Github+Plugin)
- [Javadoc Plugin](http://wiki.jenkins-ci.org/display/JENKINS/Javadoc+Plugin)
- [Jenkins Cobertura Plugin](http://wiki.jenkins-ci.org/display/JENKINS/Cobertura+Plugin)
- [Mailer](http://wiki.jenkins-ci.org/display/JENKINS/Mailer)

For email notifications to work the following SMTP server had to be
configured in Jenkins: `10.100.11.55` (all other fields left on their
default value). This is the address of `mail.erlang-solutions.com`,
but the server couldn't resolve the host name.

### Log into Jenkins server

Put this into `~/.ssh/config` and create `~/.ssh/wombat_jenkins` (see the
Jenkins access document).

```
Host ci-wombat
User jenkins
Hostname 10.100.10.123
IdentityFile ~/.ssh/wombat_jenkins
# Without this, the client would lose the connection after a few minutes. This
# is the same with all machines in the data center.
ServerAliveInterval 60
```

Now you can log in:

```
ssh ci-wombat
```

If you need root access, ask someone in the Wombat team or a sysadmin.

### Configuring a new Jenkins slave machine

#### What to save from the old machine

* Job configurations
* `~/.gitconfig`
* `~/.ssh`
* `~/.bash_history` (both `root` and `jenkins` user)
* `~/.tmux.conf`

Copy these to the new machine.

#### Config files

In the new machine, add this to the `~/.bashrc` file of the `root` and the
`jenkins` user, so that we will see which commands were executed and when (at
least when we properly quit the shell):

```sh
export HISTFILESIZE=1000000
export HISTSIZE=100000
export HISTTIMEFORMAT="%Y-%m-%d %H:%M:%S "
```

### tmux and mosh

```
apt-get install vim tmux mosh
```

`tmux` and `mosh` are useful because the connection towards the data center
machines often times out after a few minutes.

The following snippet in your `.bashrc` is quite handy when connecting to the
Jenkins machine:

```
# Usage: mot <machine> <tmux_session_name>
function mot
{
    # Last parameter -> session name
    SESSION_NAME="${@: -1}"
    #
    # Remove last parameter from "$@"
    set -- "${@:1:$(($#-1))}"

    mosh "$@" -- bash -c "tmux attach -t $SESSION_NAME || tmux new -s $SESSION_NAME"
}
```

### SSH configuration

Jenkins needs access to the private GitHub repositories used by the
Wombat project (currently: `elarm`, `deltacloud` and `wombat`
itself). This is done by adding a *deploy key* to the project.

The problem is that a single deploy key cannot be reused on multiple
repositories. But we cannot configure in git (or rebar) which key to
use for which repository directly. The workaround is that we rewrite
the host names to an alias:

- from `git@github.com:esl/PROJECT.git`
- to `git@github-PROJECT:esl/PROJECT.git`

The `jenkins` user needs to have a corresponding entry in her
`~/.ssh/config` file:

```
Host github-PROJECT
HostName       github.com
User           git
IdentityFile   /var/lib/jenkins/.ssh/PROJECT_key
IdentitiesOnly yes
```

The `PROJECT_key` files can be generated with `ssh-keygen`. At first
it asks the name of the key file, you should specify a different name
like `wombat-username`.  Hit enter and leave the passphrase empty as
well.

When the key is generated add that key to the deploy keys of your
project in `github.com`.  Go to your repository, click Setting on the
right, and choose Deploy keys in the left menu.  Choose a name like
`ESL Jenkins` and paste your public key into the text area.

The last step is to explain git to use the server alias instead of
`github.com`. Fortunately there is an option for git that allows
rewriting URLs (it was designed to e.g. replace `git://` with
`https://` protocol and alike, but it's also very handy for us).

As the `jenkins` user add an URL rewrite rule like this:

```
git config --global url.git@github-PROJECT:esl/PROJECT.git.insteadof \
                    git@github.com:esl/PROJECT.git
```

Now Jenkins can access your repository.

When setting up CI jobs for individual project members they need
similar aliasing for their personal Wombat forks. The above guidelines
shall work, simply replace the `:esl` username with the developer's
account id.

### Open issues

- The web interface displays time in UTC instead of CET which would be
  more convenient for us.
- GitHub push notifications doesn't seem to work (neither via the
  GitHub plugin, nor via the Git Plugin). This is probably because our
  server is firewalled off from GitHub. The workaround is to poll
  GitHub periodically.

## Software on the CI server

Besides Jenkins, the following programs need to be installed on Jenkins for the
builds and tests to work properly.

**Note** The steps below were written when these packages were install on a
Debian machine, but the current CI server is an Ubuntu 12.04 machine, so they
don't necessarily work exactly as described here. The reason for the server
being on a 12.04 Ubuntu machine is that there is a bug in R15B03 (which is used
to compile Wombat): when compiled with a new enough GCC, GCC wants to optimize
some C code whose behaviour is undefined. This was fixed by Lukas 3 years ago,
after R15B03. See this email thread for the details:
http://erlang.org/pipermail/erlang-bugs/2013-December/003919.html. This also
explains why we didn't see the same error in our old Jenkins machine (older GCC)
or our Macs (Apple's C compiler).

### JS test: Karma

    # As root:
    cd /usr/local/lib
    npm install karma # 1.3.0
    npm install jasmine # 2.5.2
    npm install karma-jasmine # 2.5.2

The Debian folks decided to rename the `node` command to `nodejs`, but PhantomJS
expects it to be `node`, so we create a `node` symlink:

    # As root:
    (cd /usr/bin && ln -s nodejs node)
    npm install phantomjs # 2.1.7

    npm install karma-junit-reporter # 1.1.0
    npm install karma-phantomjs-launcher

cd /usr/bin
ln -s /usr/local/lib/node_modules/phantomjs/lib/phantom/bin/phantomjs .

### JS test: PhantomJS 2.0 (not applicable any more)

    #sudo apt-get install build-essential g++ flex bison gperf ruby perl \
    #  libsqlite3-dev libfontconfig1-dev libicu-dev libfreetype6 libssl-dev \
    #  libpng-dev libjpeg-dev python libx11-dev libxext-dev
    #
    # Not needed any more
    #git clone git://github.com/ariya/phantomjs.git
    #cd phantomjs
    #git checkout 2.0
    #./build.sh
    #cp bin/phantomjs /usr/local/bin/

### JS test: Selenium

    # As root:
    apt-get install python3-pip
    pip3 install selenium

### MarkdownPP, gimli

    # As root:
    apt-get install python-pip ruby-dev

Afterwards these tools are installed automatically by `wombat-docs/check-deps.sh`.

### mkdocs

    # As root:
    apt-get install python-dev
    pip install mkdocs

### Erlang

    # As root:
    apt-get install ncurses-dev

Download https://github.com/kerl/kerl to `~/bin` and add this to `~/.bashrc`:

```
export PATH="$PATH:$HOME/bin"
export KERL_DEFAULT_INSTALL_DIR=$HOME/.kerl/installations/
```

Then:

```
kerl update releases
kerl build R16B02 R16B02 && kerl install R16B02
kerl build R15B03 R15B03 && kerl install R15B03
kerl build R14B04 R14B04 && kerl install R14B04
kerl build 18.3 18.3 && kerl install 18.3
```

The following warnings were printed during the builds, but apparently they are
fine:

```
APPLICATIONS DISABLED (See: /home/jenkins/.kerl/builds/R16B02/otp_build_R16B02.log)
 * jinterface     : No Java compiler found
 * odbc           : ODBC library - link check failed

APPLICATIONS INFORMATION (See: /home/jenkins/.kerl/builds/R16B02/otp_build_R16B02.log)
 * wx             : wxWidgets not found, wx will NOT be usable

DOCUMENTATION INFORMATION (See: /home/jenkins/.kerl/builds/R16B02/otp_build_R16B02.log)
 * documentation  :
 *                  xsltproc is missing.
 *                  fop is missing.
 *                  xmllint is missing.
 *                  The documentation can not be built.
```

Don't forget to release the builds to free up the space:

```
kerl list builds
kerl remove build R14B04
...
```

### Elixir

Install https://github.com/taylor/kiex (also follow the instruction to add a
line to the `.bashrc`).

```
. /home/jenkins/.kerl/installations/18.3/activate
kiex list known
kiex install 1.3.0
```

### Other dependencies for the tests

    # As root:
    apt-get install lsof

## Jenkins jobs

We shall set up a job for every important branch (`master` and `dev`)
and for each user contributing to the project. All jobs shall be
prefixed with `wombat-`.

*Note:* for some reason there is also a `dev-handbook` job on our
 Jenkins instance that builds the developer's handbook.

Recommended job configuration:

- Description: this is a very useful field because it lets us add
  arbitrary HTML right under the project's name on the index page. I'd
  recommend linking the build overview page from here:
  ```html
  <a href="lastSuccessfulBuild/artifact/overview.html">Wombat Overview</a>
  ```
- GitHub project: should be `https://github.com/esl/wombat`, provides
  a link to the GitHub page on the sidebar and links commits from
  changelogs to their appropriate GitHub page.
- Restrict where this project can be run: `wombat` must be selected, so that the
  jobs will run only on the Wombat slave machine (ci-03).
- Source Code Management: select Git and fill in:
  - Repository URL: `git@github.com:esl/wombat.git` (note that
    [URL rewriting](#ssh-configuration) has to be set up properly).
  - Credentials: none.
  - Branch Specifier: the name of the branch to build (for users' jobs
    it's best left empty to build every branch).
- Build Triggers: select Poll SCM (push notifications don't work). I'd
  recommend the schedule `H/15 8-20 * * 1-5` to poll GitHub every 15
  minutes in reasonable work hours.
- Build:
  - First we copy the necessary dependencies (this requires the "Copy Artifact"
    Jenkins plugin to be installed). We always copy into the `artifact_deps`
    directory, because other directories will be removed by `git clean`.
  - Then we need an "Execute shell" build step, but with the following command:

    ```bash
    scripts/jenkins.sh
    ```

Create the following post-build actions:
- Archive the artifacts; with the following (comma separated) file
  list:
  - `overview.html`: the overview page generated in the last build
    step.
  - `*.tar.gz`: the packaged Wombat release.
  - `apps/*/logs/**`: CT logs from every application.
  - `apps/*/doc/**`: EDoc pages from every application.
  - `apps/wo_dashboard/priv/js/lib/bootstrap/**`: the CSS file used by
    the overview page.
  - `rel/*/*/log/**`: log files from the generated nodes (they will be
    created in integration tests which are not yet present).
  - `rel/*/*/data/**`: data directories from the generated nodes.
  - `dialyzer_output`: Dialyzer's standard output and standard error.
  - Open the Advanced settings and check Do not fail build if
    archiving returns nothing.
- Publish Cobertura Coverage Report; with the following (comma
  separated) file list:
  - `apps/*/.eunit/TEST-*.xml`: coverage reports from EUnit tests.
  - `apps/*/logs/*/junit_report.xml`: coverage reports from CT tests.
  - Open the Advanced settings and set the desired Coverage Metric
    Targets. My recommendations:
    - Files: 90/80/0.
    - Lines: 80/50/0.
- E-mail Notification; to send emails about successful/failed builds
  to team members. I'd recommend sending an email for every unstable
  build.

By the way, the easiest way of creating new Jenkins jobs is to copy an
existing one.

### Different Jenkins jobs

**Summary**

The wombat-PLT job builds a PLT containing OTP and all deps. It is used by
the wombat-dev and other similar jobs (wombat-branch, wombat-temporary, etc),
which build Wombat, then run Dialyzer on R15B03 and all automatic tests on
R15B03.

The wombat-PLT-18 job builds a PLT containing OTP 18. It is used by
the wombat-elixir job, which builds Wombat, then re-generates the Wombat release
with OTP 18 (so that Wombat will be happy to talk to the Elixir plugins who
might send back maps), runs Elixir-specific CT tests with OTP 18, and runs
Dialyzer on Elixir plugins with OTP 18.

In case of both jobs, different parts of Wombat are compiled with different OTP
versions. (Wombat server: R15B03, Erlang plugins: R14B04, Elixir plugins: 18.3.)

**wombat-PLT**

    . /var/erlang/r15b03/activate
    git clean -f -d -x
    rm -rf deps/*
    rm -f .dialyzer_plt || true
    make compile
    make .dialyzer_plt

**wombat-PLT-18**

    . /usr/local/lib/erlang/18.3/activate

    ERLANG_DIALYZER_APPS="erts kernel stdlib compiler tools crypto ssl runtime_tools os_mon public_key inets xmerl sasl mnesia common_test edoc"

    rm -f OTP.plt || true

    # Dialyzer returns 2 in case it finds problems in the code,
    # so we should return 0 when Dialyzer returns either 0 or 2.
    dialyzer --output_plt OTP.plt --build_plt --apps $ERLANG_DIALYZER_APPS; [ $? -ne 1 ]

**wombat-elixir**

    scripts/jenkins_elixir.sh

**wombat-dev**

    scripts/jenkins.sh

### Temporary Jenkins jobs for testing pull requests

Get the `API_TOKEN` by going to Jenkins → Username → Configure → Show API token.

Then, in order to test a pull request:

```
USERNAME=wombat_dev
API_TOKEN=...
curl -X POST http://${USERNAME}:${API_TOKEN}@jenkins.erlang-solutions.com/job/wombat-temporary/buildWithParameters \
     -d PR=<PULL-REQ> -d EMAIL=<YOUR@EMAIL>
```

## Testing tools and conventions

### Private ESL repositories

If a new private repository is added to the dependencies you will have
to generate a new SSH key on the CI server, register it on GitHub and
set up [URL rewriting](#ssh-configuration).

### Executing all test cases even if some of them fails

If you have subdirectories in your `rebar.config` file and execute
`rebar eunit` or `rebar ct` the testing will stop at the first
subdirectory with failures. The workaround is to use the following
make rules:

```Makefile
test:
	$(MAKE) -k $(foreach app,$(wildcard apps/*),$(notdir $(app)).test)

%.test:
	cd apps/$*; \
	../../rebar -C ../../rebar.config -k eunit ct skip_deps=true
```

Explanation:
- We use a recursive make call to execute tests in every
  application. The `-k` switch tells make to continue building
  independent targets even after a failure (but fail at the end of the
  execution). This will guarantee that tests are executed in every
  application.
- The suffix rule calls `rebar` with `-k` and `-C` options:
  - `-k` is similar to make's switch.
  - `-C` specifies to use the *top-level* `rebar.config` file, because
    the individual applications don't have any.

However, there are some side effects os using the top-level
`rebar.config`:
- Relative pathes (e.g. `lib_dirs` and `deps_dir`) need to be updated.
- The `sub_dirs` option should not be used.

These modifications are performed by a
[rebar.config.script](../rebar.config.script), so the config is save
to use from either top-level or from within an application. (This is
true for any `rebar` target of course, not just `eunit` and `ct`.)

This whole process is quite generic and applies to newly created
applications within the Wombat project automatically.

### Code coverage measurement from tests

Generating code coverage in Cobertura format is a three-step process:
1. Turn on coverage measurement in EUnit.
2. Turn on coverage measurement in CT.
3. Convert the results to Cobertura format.

For EUnit code coverage measurement is turned on globally via
`rebar.config`. CT however requires a `cover.spec` file to be present
in every application's `test` directory:

```erlang
%% -*- erlang -*-
{import, ".eunit/eunit.coverdata"}.
{export, "logs/ct.coverdata"}.
{incl_dirs_r, [".."]}.
```

The `import` directive tells CT to start from the coverage info
gathered during EUnit tests. This would cause CT to fail if EUnit had
not been executed before it, which is important to know if someone
wants to executes the tests manually.

The final format conversion step is done by
[covertool](https://github.com/idubrov/covertool), however make
invokes it via our custom
[covertool_wrapper.es](../scripts/covertool_wrapper.es) script for a
number of reasons:
- this script picks CT coverdata whenever it exists but fails back to
  EUnit coverdata otherwise.
- `covertool` produces many `Module XY already imported from
  "filename"` warnings due to a bug.
- there's no need to escriptize covertool, which would result in
  further `rebar` warnings.

If a new application is added to the Wombat project an existing
`cover.spec` needs to be copied from an other application to its
`test` directory. This file only needs customization if there are
modules that cannot be cover compiled, e.g. `.beam`-s compiled from
`.dtl` templates or similar. These modules need to be excluded from
cover analysis in the `cover.spec`.

### Surefire reports from test results

Jenkins understands Surefire XML reports from test results. The
generation of these reports is turned on in EUnit and CT via the
`eunit_opts` and `ct_extra_params` settings in the `rebar.config`.

This feature doesn't need any maintanence when new applications are
added to the project.

### Project overview page

At the end of the build an overview HTML page is generated by `make
build_overview`. This target calls
[make_overview.es](../scripts/make_overview.es) to collect links to
the artifacts we shall keep from each build:
- EDoc pages
- CT test results (which are much more verbose than the default error
  report shown by Jenkins)
- the `.tar.gz` packages.

The page is rendered from a Mustache template and links to the
`bootstrap.css` in the `wo_dashboard` application to have a decent
look.

### Integration tests

Wombat definitely needs some integration tests that cover the
cross-application aspects of the project, like:
- Dialyzer tests
- Xref checks
- End-to-end functional and load tests.

This is however just a plan for the future at the moment.
