# Wombat Developer's Guide

## Table of Contents

1. [Developing Wombat](#developing-wombat)
2. [Overall architecture](#overall-architecture-of-wombat)
3. [Code layout](#code-layout)
4. [Important concepts and techniques](#important-concepts-and-techniques)
5. [Database upgrades](#database-upgrades)

## Developing Wombat

Please follow the [ESL Development Handbook][dh].
Especially the [coding conventions][dh-cc].
Especially the following rules:

* [Use only spaces for indentation.][dh-indent]
* [Keep line length at 80 characters.][dh-mand]
* Use the module templates of the [ESL Tempate Project][template-modules].
* [Add enough information to the commit messages and format them properly.][dh-commit-msg]
  The first line of the commit message shouldn't be more than 72 characters.
  (Otherwise it will look ugly on GitHub.)

[dh]: https://github.com/esl/dev-handbook/blob/master/development_handbook.markdown
[dh-cc]: https://github.com/esl/dev-handbook/blob/master/development_handbook.markdown#coding-conventions
[dh-mand]: https://github.com/esl/dev-handbook/blob/master/development_handbook.markdown#mandatory-rules
[dh-indent]: https://github.com/esl/dev-handbook/blob/master/development_handbook.markdown#indentation
[template-modules]: https://github.com/esl/esl-template-project/tree/master/module_templates
[dh-commit-msg]: https://github.com/esl/dev-handbook/blob/master/development_handbook.markdown#commit-messages

## CircleCI

Before a PR can be accepted, CircleCI needs to pass on it.

CircleCI access:

* URL: https://circleci.com/
* Connect with the Github account
* 
### Makefile

The beginning of the `Makefile` describes how to use it.

### Automatic tests

We have several automatic tests.

**xref**

Running xref on the whole repo:

    ./rebar xref skip_deps=true

**Dialyzer**

Running Dialyzer on the whole repo:

    make analyze

**EUnit**

EUnit is used to write and run lightweight unit tests.

Running all EUnit tests:

    make rel  # Only the wombat_plugin tests need the rel.
    make eunit

You can also run EUnit on just one application:

    export ERL_EPMD_PORT=44444
    ./rebar eunit -j1 apps=wo_core

You can even create a Git commit hook that executes the unit tests and shows
popups:

    $ brew install terminal-notifier  " notifier app for the Mac
    $ ln -s ../../scripts/pre-commit .git/hooks/pre-commit

You may want to go to System Preferences → Notifications to change the
notification type from Banner to Alert, so that the popup will disappear only if
you click on it.

**Common test**

Common test is used to write and run system tests.

Preparation before executing common test locally:

1. Ensure that Wombat is stopped.
2. Ensure that your `wombat.config` is empty (there are some settings that
   will make the test fail).
3. Build a release and compile the test plugin modules:

        make rel
        make test_plugin

Running all common tests:

    make ct

Running one test suite:

    make ct-suite SUITE=wo_rest

If you want to abort a CT test or some CT nodes weren't stopped by CT, you can
kill all CT nodes:

    make ct_kill_nodes

**Karma**

Karma is used to write and run Dashboard tests.

If you modify the web dashboard, you should also run our web dashboard unit
tests:

1. Install NodeJS from http://nodejs.org/

2. Install karma:

    $ npm install karma

3. Start karma, which will automatically run dashboard unit tests when the code
   changes:

    $ cd apps/wo_dashboard/
    $ karma start dashboard.conf.js


**Compiling from Vim**

It is possible to compile the currently opened Erlang source file in vim
and load the compiled code into the running wombat server. Set the makeprg
variable to "\<wombat-inst-dir\>/scripts/compile.sh\ %:p"
(e.g.

    set makeprg=/home/nar/wombat/scripts/compile.sh\ %:p
    
), then the
:make command will compile the current source file and load the generated
module on all nodes. The compilation errors and warnings will show up in the
quickfix window.

## The overall architecture of Wombat

**Note** What this document calls "plugins" (Metrics, Notifications, Alarms,
etc.) are now called "services". Plugins in the new sense are being implemented
now. They will be smaller and different components: a typical plugin will send a
few metrics, notifications and alarms towards the rest of Wombat. When the code
is modified according to the new terminology, this document shall be modified
too.

The main aim of Wombat is to **manage nodes**, including monitoring the metrics,
logs and alarms of nodes, and deploying new nodes.

The generic architecture of Wombat has two important aspects (a vertical and a
horizontal aspect), which arise from two important requirements: scalable
architecture and plugin infrastructure.

The overall architecture of Wombat is depicted by the following image:

<img alt="Wombat architecture" src="arch.png" width="600">

### Wombat's vertical architecture

Vertically, the Wombat architecture has the following levels, where each level
talks only to the levels below and above it:

1. **Wombat web dashboard and Wombat CLI**. They translate the user's actions
   into REST requests. (Example: when the user adds an existing Erlang node to
   Wombat by submitting a form on the Web interface, the web dashboard will send
   a `POST /action/add-node` request to the REST interface.)
2. The **Northbound interface** (a.k.a. Wombat API) currently means a REST
   interface. When it receives a request, it calls function on the master level
   to provide the answer. (Example: when it receives `POST /action/add-node`,
   it calls `wo_node_disc:add_node(Node, NodeDiscovery)`).
3. The **master processes** are central processes that are global to the Wombat
   system. They manage and use the middle managers. There is one core master
   process and one master process for each plugin. The core master process
   manages not only the core middle manager processes, but also the plugin
   master processes. (Example of a request handled by the core master process:
   when `wo_node_disc:add_node(NodeName, NodeDiscovery)` is called on the
   master, it will in turn call `wo_node_mgr_master:do_manager_node` (still on
   the master), which decides which middle manager to use and calls
   `wo_node_mgr_mm:manage_node(MiddleManagerId, NodeId)` to ask the appropriate
   middle manager to start managing the node.
4. A **middle manager** (abbreviated as "MM") is  set of processes that manage a
   number of Erlang nodes. A middle manager has a core middle manager process
   and one middle manager process for each plugin. The core middle manager
   process manages the plugin middle manager processes. We introduced middle
   managers for scalability reasons: if we have many Erlang nodes, we may need
   to divide them among several middle managers. When
   `wo_node_mgr_mm:manage_node(MiddleManagerId, NodeId)` is called, the middle
   manager connects to the given node with the given cookie. Afterwards a
   notification is sent to all subscribers (which are usually other subsystems),
   so they can prepare for managing the new node (e.g. by starting new
   processes). When the subsystems have finished their preparation, the plugin
   infrastructure injects the plugin modules and starts the plugin processes
   (a.k.a. agent processes).
5. **Agent processes** run on the Erlang nodes of the managed system. They are
   injected by Wombat's plugin infrastructure (the `wo_plugins` application). We
   are writing them to be as non-intrusive as possible.

**Note** The separate master and MM level were introduced because we wanted to
build networks of Wombat nodes by having one master node running the master
processes and several MM nodes each running a set of MM processes. But after
when we got to start scaling to multiple nodes, another approach was chosen
called "Meta Wombat".

### Wombat's horizontal architecture

Horizontally, Wombat has a core, and a number of subsystems. Subsystems include
the Metrics subsystem, the Notification subsystem, the Alarms subsystem and the
Orchestration subsystem (and there are several smaller subsystems). A subsystem
must provide at least [two callback modules](../new-subsystem.md) – one
implementing the master process of the subsystem, and one implementing the
middle manager process of a subsystem:

1. The subsystem's master process is started by the core master process, and it
   can use Wombat's libraries to extend Wombat's REST interface (e.g. the
   metrics subsystem adds a REST handler module to the REST level, which will
   handle the requests whose URL starts with `/metrics`).
2. The core middle manager process starts the subsystem's middle managers
   process.

This way, a subsystem will usually have a portion in most levels: it has its own
REST interface, its own master process, and its own middle manager processes.

When Wombat is started, the subsystems are started one by one. (Currently the
list of subsystems is hardcoded in `wo_core_master_sup` and `wo_core_mm_sup`.)
For each subsystem, first the master is started, then the middle manager. The
timeout in both cases is `infinity`, because that's the timeout value used by
supervision trees, and the subsystem processes themselves might implement
supervision trees, so Wombat would produce problematic behaviour if the process
above those subsystem supervision trees would have finite timeouts.

Subsystems (and any other Wombat code) can use events to get notified about new
nodes, nodes going down, etc. The `wo_event` module implements a
publish/subscribe library (it is actually just a thin wrapper around `gproc`).
Services like `wo_topo` add their own layer on the top of `wo_event` to make
sending notifications and subscribing to them more convenient for other
processes.

### Summary of the vertical and horizontal architectures ###

The **vertical architecture** of Wombat helps **scalability**, because we can
start as much middle managers to manage the nodes as needed, and the middle
manager processes can filter information to the master processes. The
**horizontal architecture** of Wombat provides an **extensible infrasctucture**.

Notes:

* The Northbound interface can have multiple flavors (SNMP, REST,
  HTTP/Websockets, ...). Not all these flavors will be implemented by all the
  services (e.g. the Orchestration service will only provide a REST flavor).
* Northbound API is used to add/enable/disable O&M services. Northbound API
  (e.g. REST) only used by external consumers (e.g. Web-based dashboard,
  existing O&M front-ends, batch jobs, ...).
* Inter-service communication happens atop the internal (Erlang) API.
* Information common to all services (e.g. node families, nodes, ...) is kept in
  a common data-base (Core DB) all services have access to.
* The middle manager has direct access to the Erlang node where the Erlang
  application is running because the node where a middle manager is running has
  been connected to it as a hidden node. This is not true in the Orchestration
  service, which only has 1) HTTP/REST access to the infrastructure provider API
  (through Deltacloud), and 2) SSH access to the deployed Virtual Machines (VMs)
  hosting the Erlang nodes.
* O&M-related services (Metrics, Logging, etc.) share a common infrastructure
  (e.g. aggregation and management layers). A O&M service must implement the
  following components:

    * North-bound interface (possibly several of them)
    * Internal (Erlang) API (inter-service communication)
    * Specialized aggregation layer
    * Specialized management layer
    * Specialized agent

## Code layout

### Releases

Wombat consists of two releases (they can be found in the `rel` directory):

* `wombat`: The Wombat server, including the REST interface.
* `wo_test`: A tiny test application used only for testing Wombat.

### Applications

Wombat consists of several applications (they can be found in the `apps`
directory). Most of them are part of the `wombat` release.

* `wo_utils`: Common headers and  utility functions.
* `wo_core`: The core of Wombat, which implements the following functionality:

  * A common database (`wo_core_db`).
  * The topology database (`wo_topo`) and REST handler.
  * The Node Manager subsystem, which manages connections with the nodes
    including node monitoring and node discovery.
  * A helper module for `cowboy` (`wo_cowboy`).

* `wo_core_master`: Starts and supervises the master components of subsystems.
* `wo_core_mm`: Starts and supervises the middle manager components of
  subsystems.
* `wo_rest`: This application implements a generic REST interface, into which
  other applications can plug in their own REST interfaces (i.e. they register
  their `cowboy` dispatch rules and handler modules) via the `wo_cowboy` module.

Subsystem applications:

* `wo_custom_cmd`: The Custom Command subsystem, which allows users to register
  and call custom Erlang functions on the managed Erlang nodes.
* `wo_metrics`: The Metrics subsystem, which collects, stores and displays
  metrics collected on the managed Erlang nodes.
* `wo_orch`: The Orchestration subsystem, which can deploy Erlang nodes in the
  cloud, and afterwards hands them over to the rest of Wombat.
* Note: The Node Manager and the topology components (see above) could also be
  subsystems, but we didn't want to increase the number of applications
  unnecessarily, so we kept them as part of the `wo_core` application.

Other applications:

* `wo_dashboard`: Application that implements the web dashboard. It contains
  static files (`index.html`, HTML templates, JavaScript files, CSS files) and a
  simple web server that serves these pages using `cowboy`. The dynamic nature
  of the web dashboard is accomplished by using the AngularJS framework, which
  is a client-side MVC framework.
* `wo_test`: A tiny test application that constitutes the `wo_test` release.

### Packaging

The [official Wombat package][installation] is with a
[process][release-process] that relies on [Jenkins to build the
package][ci-job] via the `deliver` make target, thus the
`scripts/make_release.sh` script.

There are other scripts for packaging Wombat:

* `scripts/make_orch_release.sh`: Create an OTP release of Wombat
  suitable for deploying with Wombat orchestration. Not used for
  anything particular at the moment;
* `scripts/make_otp_release.sh`: Pack the regular OTP release of
  Wombat. Unused and superseded by `scripts/make_release.sh`;

[installation]: https://github.com/esl/wombat/blob/bbe3998d1229ad6ba804c729fe4f91a6b4513e84/README.md#installing-from-a-package
[release-process]: https://github.com/esl/wombat/blob/bbe3998d1229ad6ba804c729fe4f91a6b4513e84/doc/release_notes/release_process.md
[ci-job]: https://github.com/esl/wombat/blob/bbe3998d1229ad6ba804c729fe4f91a6b4513e84/doc/ci.md#jenkins-jobs

## Important concepts and techniques

### UUIDs: strings vs binaries

In Wombat, the objects (node families, nodes, etc.) that are managed and often
even created by the user are identified with UUIDs. When a request comes in to
create an object, Wombat will generate a unique UUID, create the object, and
send the UUID as part of the response, so the caller can refer to the object in
the future. The objects themselves are Erlang records that are stored in a
Mnesia database (hidden behind modules like `wo_topo` and `wo_mnesia`).

For example node ids are UUIDs. Originally they were stored everywhere as
strings (see the `wo_topo:node_id()` type). Storing 40 string characters needs
least 40 * 9 = 360 bytes (because there is a pointer after each character), even
when the string is stored in an ETS table. Binaries are much better for storing
UUIDs, so several module use binaries instead (see the `wo_topo:node_id_bin()`
type).

## Database upgrades

Wombat needs to remain backward compatible with data files created in
earlier versions of the system. This chapter gives some guidelines for
achieving this goal.

### Mnesia tables

Most of the data is in a Mnesia database. All schema operations (like
creating tables) must go through the `wo_mnesia` module.

The tables created by `wo_mnesia:ensure_table/4` have an associated
version number (an integer value starting from 1 and incremented each
time the table's format is modified). When creating a table the
current version number and an upgrade function must be specified. If
an older version of the table exists in Mnesia `wo_mnesia` will take
care of transforming it to the requested version with the upgrade
function.

Therefore whenever you change a table in a non-backward-compatible
way:

1. increment the version number; and
2. if the table was using `wo_mnesia:default_upgrade_table/2` as an
   update function, write a new, custom upgrade function for the table;
3. otherwise add a new clause to the custom upgrade function that can
   transform the __data in the table__ and the __properties of the
   table__ to the new format.

The upgrade function is _incremental_, which means it should be
capable of upgrading a table of version _N_ to _N+1_ in one step. If
the actual table is multiple versions behind the version requested by
the code, `wo_mnesia` will call the upgrade function multiple times.

It is important to state that the _properties of the table_
(e.g. type, index fields, record name etc.) need to be upgraded too,
otherwise `wo_mnesia:ensure_table/4` will fail. In many cases it's not
possible to modify a table's properties after it is created. In that
case you need to delete and recreate the table in the upgrade function
(and of course saving the values to a temporary place in the
meantime).

### Custom data files

There are a number of custom data files used by Wombat
(e.g. `disk_log` files for notifications or `wo_metrics_seq_file` for
the metrics data). These file formats also need custom upgrade code.

The preferred solution is to follow the _upgrade on read_ pattern,
which works well for files containing multiple records:

* Always write records in the most recent format to the file.
* Convert data read back from the file immediately to the most recent
  record format with an upgrade function.

If the changes in the file format are too drastic to keep old an new
format records in the same file, use the _upgrade to new file_
pattern:

* Data in new format shall be stored in separate files from data in
  the old format.
* The old and new format files are distinguishable by their naming
  convention.
* When an old format file is found at start up, copy data from it to a
  new format file and then delete the old format file.

#### Best practices for upgrade on read

There should be a central upgrade function for data records in the
module handling the data files.  When a new record format is
introduced, this upgrade function needs to be updated as well.

* Add a new clause to the _end of the function_ that matches the new
  record format, and returns it unmodified.
* Do _not_ use record notation in the match pattern, but spell out the
  entire tuple. It is advised to add type specs as comments to the
  fields.

  Record definitions and type specs may change from version to version
  of the data, the goal here is to document one specific version's
  characteristics.
* If you get a _this clause cannot match because a previous clause at
  line N always matches_ compiler error than the new format cannot be
  distinguished from a previous format. Fix that by either modifying
  the new format or adding guards to the previous clause.
* Ensure that all previous clauses upgrade the data to the newly
  introduced format. Ideally this could be done by modifying the
  previous clause only, like in this example:

  Before:
  ```erlang
  upgrade_record_format(V1Pattern) ->
      V2 = upgrade_v1_to_v2(V1Pattern),
      upgrade_record_format(V2);
  upgrade_record_format(V2Pattern) ->
      V3 = upgrade_v2_to_v3(V2Pattern),
      upgrade_record_format(V3);
  ...
  upgrade_record_format(VNPattern) ->
      VNPattern.
  ```

  After:
  ```erlang
  upgrade_record_format(V1Pattern) ->
      V2 = upgrade_v1_to_v2(V1Pattern),
      upgrade_record_format(V2);
  upgrade_record_format(V2Pattern) ->
      V3 = upgrade_v2_to_v3(V2Pattern),
      upgrade_record_format(V3);
  ...
  upgrade_record_format(VNPattern) ->
      VNplus1 = upgrade_vN_to_vNplus1(VNPattern),
      upgrade_record_format(VNplus1);
  upgrade_record_format(VNplus1Pattern) ->
      VNplus1Pattern.
  ```

#### Best practices for upgrade to new file

In this upgrade scenario it is very important not to lose the data
stored in the old file if something goes wrong during the upgrade. So
it is best to use a three-file pattern:

1. The data from the old file is copied to a temporary file (also
   distinguishable by its name) in the new format.
2. The old file is deleted.
3. The temporary file is renamed to the new file naming schema.

Then the pseudo code for the upgrade algorithm may look like this:
```erlang
upgrade_to_new_file() ->
    case {old_file_exists(), temporary_file_exists()} of
        {true, false} ->
            do_upgrade_from_copy_step();
        {true, true} ->
            delete_temporary_file(),
            do_upgrade_from_copy_step();
        {false, true} ->
            do_upgrade_from_rename_step();
        {false, false} ->
            %% Nothing to upgrade
            ok
    end.
do_upgrade_from_copy_step() ->
    copy_old_file_to_temp_file(),
    delete_old_file().
do_upgrade_from_rename_step() ->
    rename_temp_file_to_new_file().
```

[issue-448]: https://github.com/esl/wombat/issues/448
[dispatch-rule]: https://github.com/esl/wombat/blob/d7a99ba44d5d82e084924b34e17796de09d4c502/apps/wo_metrics/src/wo_metrics_master.erl#L89
[inject-agent]: https://github.com/esl/wombat/blob/d7a99ba44d5d82e084924b34e17796de09d4c502/apps/wo_metrics/src/wo_metrics_mm.erl#L394
