## How to write a new Wombat subsystem?

**Note** Subsystems (e.g. Metrics) used to be called "plugins", and this is
still reflected in some of the code (e.g. subsystems are started with the
`start_link_plugin` callback). Now by "plugin" we mean an agent process that
sends a specific kind of information to Wombat (e.g.
`wombat_plugin_builtin_memory_metrics`).

### Callback modules and functions

You need to create an Erlang application with at least two files: a master
callback module and a middle manager callback module.

#### Master callback module

`apps/wo_<subsystem_name>/src/wo_<subsystem_name>_master.erl` needs to implement
the callback module which specifies how to start the **subsystem master process**.

It has the following mandatory callback functions:

* `start_link_plugin/0`: Starts the subsystem. This function should either start
  a `gen_server`-like process (and behave like `gen_server:start_link`) or
  start a supervision tree (and behave like `supervisor:start_link`).

  ```erlang
  start_link_plugin() -> Result

      Result = {ok, pid()} | ignore | {error, Error}
      Error = {already_started, Pid} | term()
  ```

* `config/0`: Returns config information. Currently it only returns whether the
  process that `start_link_plugin` starts is a worker (e.g. a `gen_server`) or a
  supervisor process:

  ```erlang
  config() -> [{process_type, ProcessType}]

      ProcessType = worker | supervisor
  ```

#### Middle manager callback module

`apps/wo_<subsystem_name>/src/wo_<subsystem_name>_mm.erl` needs to implement the
callback module which specified how to start the **subsystem the middle manager
process**. It has very similar callbacks to the master callback module above,
except that it also gets the id of the middle manager. Currently we always have
one middle manager called `"default_mm"`, but in the future we will have any
number of middle managers in a Wombat instance.

`wo_<subsystem_name>_mm.erl` has the following mandatory callback functions:

* `start_link_plugin/1`: Starts the subsystem on a middle manager. This function
  should either start a `gen_server`-like process (and behave like
  `gen_server:start_link`) or start a supervision tree (and behave like
  `supervisor:start_link`).

  ```erlang
  start_link_plugin(MMServerId) -> Result

      MMServerId = uuid() | "default_mm"
      uuid() = string()
      Result = {ok, pid()} | ignore | {error, Error}
      Error = {already_started, Pid} | term()
  ```

* `config/0`: Returns config information. Currently it only returns whether the
  process that `start_link_plugin` starts is a worker (e.g. a `gen_server`) or a
  supervisor process:

  ```erlang
  config() -> [{process_type, ProcessType}]

      ProcessType = worker | supervisor
  ```

#### Starting the subsystems

The subsystems and their order is hard-coded in the `wo_core` application, so if
you create a new subsystem, you just need to make sure that the two callback
modules (`<subsystemname>_master.erl` and `<subsystemname>_mm.erl`) are in an
application that is included in the wombat release, start the master process [in
`wo_core_master`][start_subsystem_master] and start the MM process [in
`wo_core_mm`][start_subsystem_mm]. (It is fine to start only one of them.)

#### Events

*Node events*

* How to subscribe:

  ```erlang
  subscribe(node_event, all)
  subscribe(node_event, {mm_id, MMId})
  ```

* How to send an event:

  ```erlang
  notify_node_event(EventType, MMId, NodeId)
  ```

* The most important messages:

  * `{node_up, NodeId}`: occurs when Wombat connects to a node (either because
    it was newly added or because it was down).
  * `{node_down, NodeId}`: occurs when a node becomes unreachable from Wombat.
  * `{node_added, NodeId}`: occurs when a node is added to Wombat. This does not
    mean that the node is available.
  * `{node_initialized, NodeId}`: occurs when a node goes from UNINIT state to
    STOPPED state.
  * `{node_deleted, NodeId}`: occurs when the user deletes a node from Wombat.
  * `{node_stopped, NodeId}`: occurs when the user stops a node from Wombat
    (only for nodes deployed by Wombat Orchestration).

* Types:

  ```erlang
  NodeId = wo_topo:node_id()
  ```

### Libraries

A subsystem might do the following things:

1. It might use the `wo_mnesia` module to store data in Mnesia tables. This data
   is persistent, so if you want to keep something even when Wombat is
   restarted, this is the best database to use. The metrics subsystem uses `mnesia`
   directly.

2. It might want to provide a REST interface. You do that by implementing a rest
   handler (see `wo_metrics_rest_handler.erl`) and [adding it to Wombat's in your
   master module][dispatch-rule].

### Supervisor tree (somewhat outdated)

The supervisor tree is shown by the following drawing:

![Supervisor tree](developers_guide/supervisor_tree.png)

Notes:

* The processes in the supervisor tree are started sequentially. A process may
  call another process synchronously (e.g. via a `gen_server:call`) only if the
  other process was started before the caller.
* The diagram does *not* reflect the order in which the processes are started.
  You can get that knowledge by looking at the code (supervisor `init`
  functions, `wo_core_(mm|master):start_plugin` functions and
  `wo_<plugin>_(mm|master):start_plugin` functions.).

We need two supervisors (a main supervisor and a plugin supervisor) on both
the master and the middle manager levels because this way we can start the
system in the following way:

1. Main supervisor is started.
2. The main supervisor starts the plugin supervisor without children.
3. The main supervisor starts the core process (`wo_core_master`/`wo_core_mm`).
4. For each plugin, the core process asks the plugin supervisor to start the
   master process for that plugin. If we had only the main supervisor, this
   would cause a deadlock, since the main supervisor is still in the process of
   starting its children (currently its waiting for the core process to return
   from its `init` function).

[start_subsystem_master]: https://github.com/esl/wombat/blob/1cbce5f8c1b30726f0a4852b44cf202241e82762/apps/wo_core_master/src/wo_core_master_sup.erl#L37-L39
[start_subsystem_mm]: https://github.com/esl/wombat/blob/1cbce5f8c1b30726f0a4852b44cf202241e82762/apps/wo_core_mm/src/wo_core_mm_sup.erl#L35-L40
