# Wombat System Testing

## Table of Contents

1. [Definition](#definition)
2. [Setup](#setup)
3. [Components: Topology](#components-topology)
4. [Components: Metrics](#components-metrics)
5. [Components: Notifications](#components-notifications)
6. [Components: Alarms](#components-alarms)

## Definition
This document defines possible system tests usage scenarios per architectural component.<br />
Some of them could be reproduced in common test suites and should be a starting point for a solid regression testing approach.

## Setup

The tests can be performed on either a developer's own Wombat build or
on Wombat installed from a package built by Jenkins. The latter
approach is favored, however it still needs an up to date copy of the
Wombat repository, because the `test-env` directory and `wo_test`
release are not part of the install package.

1. Get the up-to-date Wombat source and build the `wo_test` release:

   ```
   [ -d ~/wombat ] || git clone git@github.com:esl/wombat.git ~/wombat
   cd ~/wombat
   git fetch origin
   git checkout origin/master
   make compile wo_test.generate
   ```

   Whenever the test description says to execute a script in the
   `test-env` directory, the task needs to be performed on this
   repository.

2. Get the Wombat install package

   1. Remove old Wombat installations

      ```
      cd ~/Downloads
      rm -rf wombat-*
      ```
   2. Go to http://ci-wombat.erlang-solutions.com/job/wombat-dev/lastStableBuild/artifact/overview.html
   3. Download the `wombat-VERSION.tar.gz` file (the one without
      `-orch` in the file name).
   4. Extract, install and start Wombat:

      ```
      cd ~/Downloads
      tar zxf wombat-*.tar.gz
      cd wombat-*
      ./generate.sh
      ./start.sh
      ```

   Whenever the test description says to do something with the Wombat
   release, the task needs to be performed on this installation.

## Components: Topology

1. Start 4 erlang nodes sharing the same cookie on the local machine:

	```
	test-env/go.sh wo_test_1@127.0.0.1 cookietest;
	test-env/go.sh wo_test_2@127.0.0.1 cookietest;
	test-env/go.sh wo_test_3@127.0.0.1 cookietest;
	test-env/go.sh wo_test_4@127.0.0.1 cookietest
	```

2. Cmd: Topology + Add Node, Expect: Failure, no nodes added

	```
	name: wo_test_0@127.0.0.1
	cookie: cookietest
	discovery: false
	```

3. Cmd: Topology + Add Node, Expect: Success, nodes added `(wo_test_1@127.0.0.1)`

	```
	name: wo_test_1@127.0.0.1
	cookie: cookietest
	discovery: false
	```

4. Cmd: Topology + Add Node, Expect: Failure, no nodes added

	```
	name: wo_test_1@127.0.0.1
	cookie: cookietestwrong
	discovery: false
	```

5. Join node wo_test_3 to node wo_test_2

	```
	test-env/join.sh wo_test_3@127.0.0.1 wo_test_2@127.0.0.1 cookietest
	```

6. Cmd: Topology + Add Node, Expect: Success, nodes added `(wo_test_2@127.0.0.1, wo_test_3@127.0.0.1)`

	```
	name: wo_test_2@127.0.0.1
	cookie: cookietest
	discovery: true
	```

7. Join node wo_test_4 to node wo_test_3

	```
	test-env/join.sh wo_test_4@127.0.0.1 wo_test_3@127.0.0.1 cookietest
	```

8. Cmd: Topology + Add Node, Expect: Success, nodes added `(wo_test_4@127.0.0.1)`

	```
	name: wo_test_3@127.0.0.1
	cookie: cookietest
	discovery: true
	```

9. Join node wo_test_4 to node wo_test_1

	```
	test-env/join.sh wo_test_4@127.0.0.1 wo_test_1@127.0.0.1 cookietest
	```

10. Cmd: Topology + View Node Info + [NodeFamily] + Remove Node Family, Expect: Success, node family removed

11. Cmd: Topology + Add Node, Expect: Success, nodes added `(wo_test_1@127.0.0.1, wo_test_2@127.0.0.1, wo_test_3@127.0.0.1, wo_test_4@127.0.0.1)`

	```
	name: wo_test_1@127.0.0.1
	cookie: cookietest
	discovery: true
	```
12. Stop erlang nodes

	```
	test-env/stop.sh wo_test_1@127.0.0.1 cookietest;
	test-env/stop.sh wo_test_2@127.0.0.1 cookietest;
	test-env/stop.sh wo_test_3@127.0.0.1 cookietest;
	test-env/stop.sh wo_test_4@127.0.0.1 cookietest
	```

13. Start and stop the erlang nodes several times with the following commands.  Expect: node states are correctly shown in Topology/View Node Info

	```
	start:
	for i in {1..4}; do test-env/go.sh wo_test_$i@127.0.0.1 cookietest; done
	stop:
	for i in {1..4}; do test-env/stop.sh wo_test_$i@127.0.0.1 cookietest; done
	```

### Test how many monitored nodes can wombat handle

1. Change the license key wombat uses. With the new license added we can add a
large amount of monitored nodes.

    ```
    - Navigate to apps/wo_license directory within your
    wombat installation directory:

    cd apps/wo_license

    - Generate a new license:

    confidential/create_license.es confidential/unlimited.lic

    - Navigate back to wombat's install directory and add the license to the
    config file of wombat:

    vim rel/wombat/files/user.config

    The file should contain the generated license, like in the example below:

    [
         %% $INSTRUCTIONS$
         %%
         %% Example:
         %% {wo_rest,
         %%  [{rest_api_port, 8000}
         %%  ]}
        {wo_license, [
            {license, <the generated license within quotes>}
        ]}
    ].
    
    - Restart wombat, so the new licence will take effect.
    ```

2. Start 400 erlang nodes with the following command:

    ```
    for i in wo{1..100}; do test-env/go.sh $i@127.0.0.1 cookie; done
    ```

3. Connect the started erlang nodes:

    ```
    for i in {2..400}; do test-env/join.sh wo1@127.0.0.1 wo$i@127.0.0.1 cookie; done
    ```

4. Navigate to Topology/Add node in wombat and add node wo1@127.0.0.1, with
cookie: cookie. Use the Discover connected nodes option. Expect: All the 400
hundred erlang nodes will be added to wombat within about 20 seconds.

5. Add the wombat node itself to the list of monitored nodes:

    ```
    On Topology/Add node:
    - Node name: wombat@127.0.0.1
    - Cookie: wombat
    - Discovery: true
    ```

6. Wait some minutes and choose Metrics/Memory/Total memory, and the added
wombat node. Expect: the memory usage should not be larger than 400 megabytes.

7. Similarly choose Metrics/Runtime/CPU load for 1 avg, and select the wombat
node. Expect: the processor usage should not be more than 30%.

## Components: Metrics

1. Start 4 erlang nodes sharing the same cookie on the local machine, and connect them to each other:

	```
	for i in {1..4}; do test-env/go.sh wo_test_$i@127.0.0.1 cookietest; done;
	for i in {2..4}; do test-env/join.sh wo_test_1@127.0.0.1 wo_test_$i@127.0.0.1 cookietest; done
	```

2. Cmd: Topology + Add Node, Expect: Success, nodes added `(wo_test_1@127.0.0.1, wo_test_2@127.0.0.1, wo_test_3@127.0.0.1, wo_test_4@127.0.0.1)`

	```
	name: wo_test_1@127.0.0.1
	cookie: cookietest
	discovery: true
	```

### View Metrics

1. Cmd: Metrics + Numeric Metrics + wo_test_1@127.0.0.1 + Default + Memory + Total Memory

2. Expect:
	- single metric graph
	- navigation per different time frames
	- possible to zoom into time ranges and reset the zooming

3. Cmd: Metrics + Numeric Metrics + [NodeFamily] + Default + Memory + Total Memory

4. Expect:
	- stacked metric graph
	- navigation per different time frames
	- possible to zoom into time ranges and reset the zooming

5. Repeat steps 1-4 for all the different metrics on wo_test_1@127.0.0.1. Expect: metric graphs for all the metrics.

### Overimpose Metrics

1. Cmd: Metrics + Numeric Metrics + wo_test_1@127.0.0.1 + Default + Memory + Total Memory + Pin Metric

2. Cmd: Metrics + Numeric Metrics + wo_test_2@127.0.0.1 + Default + Memory + Total Memory

3. Cmd: Metrics + Numeric Metrics + wo_test_3@127.0.0.1 + Default + Memory + Total Memory

4. Cmd: Metrics + Numeric Metrics + wo_test_4@127.0.0.1 + Default + Memory + Total Memory

5. Expect:
	- single metric combined graph (4 nodes)
	- navigation per different time frames
	- possible to zoom into time ranges and reset the zooming
	- possible to remove node metrics using the configure metrics button or clicking on the graph legend elements

### Graphite Push Metrics

1. Start a graphite-0.9.12 instance on the testing box. Wombat should be configured for sending UDP data to graphite.

2. Open graphite dashboard `graphite-web-host:graphite-web-port/dashboard`

3. Check dashboard pattern `wombat.nodes.wo_test_*.total_memory`

4. Merge the graphs and compare to the wombat overimposed metrics

5. Expect: same dataset on both dashboards

## Components: Notifications

1. Start new test node from the shell.

	```
	test-env/go.sh wo_test_1@127.0.0.1 cookietest
	```

2. Navigate to Notifications and choose wo_test_1@127.0.0.1 node.

3. Run the following commands:

	```
	rel/common/nodetool -name wo_test_1@127.0.0.1 -setcookie cookietest rpcterms wo_test_crash_sup new_child ""
	rel/common/nodetool -name wo_test_1@127.0.0.1 -setcookie cookietest rpcterms wo_test_crash_sup crash_child ""
	```

4. Expect: In wombat notice three new notifications appearing after the second command at the
top-right corner.

5. Run these commands from shell:

   ```
   rel/common/nodetool -name wo_test_1@127.0.0.1 -setcookie cookietest rpcterms lager log 'error. test. "lager_error\n". [].'
   ```
   Expect: new notification in wombat: `lager_error`

   ```
   rel/common/nodetool -name wo_test_1@127.0.0.1 -setcookie cookietest rpcterms lager log 'critical. test. "lager_critical\n". [].'
   ```
   Expect: new notification in wombat: `lager_critical`

   ```
   rel/common/nodetool -name wo_test_1@127.0.0.1 -setcookie cookietest rpcterms lager log 'alert. test. "lager_alert\n". [].'
   ```
   Expect: new notification in wombat: `lager_critical`

   ```
   rel/common/nodetool -name wo_test_1@127.0.0.1 -setcookie cookietest rpcterms error_logger error_msg '"error_msg1\n". [].'
   ```
   Expect: new notification in wombat: `error_msg1`

6. Create an alarm on the test node with the command below. Expect: new alarm
created with Alarm id: alarm1 and Additional information: alarm1

	```
	rel/common/nodetool -name wombat@127.0.0.1 -setcookie wombat rpcterms wo_alarms_lib raise "alarm1. auto. 'wo_test_1@127.0.0.1'. node_level. \"alarm1\"."
	```

## Components: Alarms

### Stopping, and restarting the nodes

1. Start a test node from shell.

	```
	test-env/go.sh wo_test_1@127.0.0.1 cookietest
	```


2. On Topology / Add Node add a new node to wombat

	```
	Node Name : wo_test_1@127.0.0.1
	Cookie: cookietest
	Discovery: false
	```

3. On the Alarms tab select the node family which contains the node just
created. Kill the node from the shell

	```
	test-env/stop.sh wo_test_1@127.0.0.1 cookietest
	```

4. Check what happened in wombat on the Alarms tab. A red notification must
appear immediately at the top-right corner. Select the node now. Expect: an alarm
should be seen with a node_down message.

5. Start the node again:

	```
	test-env/go.sh wo_test_1@127.0.0.1 cookietest
	```

6. In wombat wait half a minute, till wombat is able to detect the node is back.
   Expect: the notification is cleared.

### Alarms generated on the managed nodes.

1. Start a test node from shell.

	```
	test-env/go.sh wo_test_1@127.0.0.1 cookietest
	```


2. On Topology / Add Node add a new node to wombat

	```
	Node Name : wo_test_1@127.0.0.1
	Cookie: cookietest
	Discovery: false
	```
3. From the shell raise an alarm on the test node:

	```
	rel/common/nodetool -name wombat@127.0.0.1 -setcookie wombat rpcterms wo_alarms_lib raise 'message1, auto, \'wo_test_1@127.0.0.1\', node_level, "alarmooo"'
	```

4. Select the node on the alarm tab. Expect: the alarm with the generated id:
message1 should appear after refreshing the page.

5. Generate one more alarm with the same parameters from the shell.

6. The new alarms should be shown in the list of alarms.


## Components: Orchestration

### Create local nodes

1. Before starting to deploy new nodes check whether the ssh key used by wombat is accepted on your machine.

	```
    - Go to your wombat installation directory, and connect to the running wombat node with:

		rel/wombat/wombat/bin/wombat attach
	
	- In the erlang shell use this command: 
    
      wo_orch_conf:private_key(). 
      
      It will return the private key used by wombat when connecting to other hosts.
	  Check whether with this key you can ssh to your host:

	  ssh -i <Result of the command in the erlang shell>.pub 127.0.0.1

	- If in the previous step ssh asked for a password you need to add the public key to your ~/.ssh/authorized_keys.
	```

2. Go to Orchestration / Infrastructure providers and click on New Provider

	```
	Fill the fields as following:
	Infrastructure provider: Physical Provider
	name: riak_prov
	servers: 127.0.0.1
	```

3. On the Releases tab fill the fields with these values:

	```
	Name: riak_rel
	Description: riak
	Start command: bin/riak start
	Stop command: bin/riak stop
	Erlang cookie: riak
	VM args file: etc/vm.args
	Release tarball: Choose given riak release tarball file depending on your os
	```

4. On the Node Families tab:

	```
	Name: riak_family
	Release: riak_rel
	Click on Add Domain button
	Infrastructure provider: riak_prov
	SSH user: your username on localhost
	Virtual Disk: ami-0cdf4965
	Hardware profile: m1.medium
	Click again on the Add Domain button
	Firewall Rules: leave empty
	Node selection strategy: Random
	Bootstrapping strategy: Custom
	Bootstrapping options: cmd:bin/bootstrap
	```

5. In the last step we deploy the nodes on the Nodes tab

	```
	Click on the Deploy Nodes button
	Node Family: riak_family
	Amount: 3
	Automatically start node(s) after deployment: true
	Click on Deploy
	```

6. After about a half a minute check whether the nodes are running by refreshing the page. Expect: Success, which is indicated by the status of the nodes.

### Deploying nodes on Amazon

1. Go to Orchestration and choose Infrastructure Providers. Fill the fields with the following values:

	```
	Infrastructure Provider: Amazon EC2
	Name: am_riak_prov
	Username:  Use your amazon username
	Password: Use your password on amazon
	```

2. On the Releases click on the New Releases button

	```
	Name: am_riak_rel
	Description: riak
	Start command: bin/riak start
	Stop command: bin/riak stop
	Erlang cookie: riak
	VM args file: etc/vm.args
	Release tarball: etc/vm.args
	Release tarball: riak_ubuntu.tar.gz
	```

3. Node Families / New Family:

	```
	Name: am_riak_family
	Release: am_riak_rel
	Click on Add Domain button
	Infrastructure Provider: am_riak_prov
	SSH User: laszlototh
	Virtual Disk: ami-0cdf4965
	Hardware profile: m1.medium
	Click on Add Domain button
	Firewall rules: 8098:8099,tcp:8087:8087,tcp:8069:8069,tcp:6000:7999
	Node selection strategy: Random
	Bootstrapping strategy: Custom
	Bootstrapping options: cmd:bin/bootstrap
	```

4. Nodes / Deploy Nodes:

	```
	Node Family: am_riak_family
	Amount: 2
	Automatically start node(s) after deployment: true
	```

5. After clicking on the Deploy button wait about 2-3 minutes. Expect: Nodes
should be created in the amazon cluster indicated by their status in wombat.

6. Remove the nodes in wombat:

    ```
    Choose Topology/View Node info, select the node you want to remove and click
    on Remove node.
    ```

7. Stop the running amazon host on the EC2 Management Console:

    ```
    Login to your amazon Management Console, select the node and choose
    actions/stop.
    ```

