#!/usr/bin/env escript
%% -*- erlang -*-
%%! -s inets start -name change_node_name@127.0.0.1

-include_lib("kernel/include/file.hrl").

-define(W(Fmt, Args), io:format(Fmt, Args)).

-define(BACKUP_NAME, "mnesia_backup").

main([]) ->
    ?W("~nUsage: ~s wombat@old.name wombat@new.name~n~n"
       "This script dumps Wombat's database to a file and renames Wombat's nodename in it.~n"
       "Wombat must be running during the operation.~n"
       "To restore the db on a fresh Wombat instance use the complementary restore_backup.es script.~n"
       "The old Wombat and the new Wombat instance must have the same version.~n~n"
       "For example: if you want to change a running Wombat's nodename from wombat@127.0.0.1~n"
       "to 10.0.0.1 run the following command while the old Wombat is running:~n"
       "~s wombat@127.0.0.1 wombat@10.0.0.1 ~n"
       "The script will print the location of the backup file it created.~n"
       "Install a new Wombat instance of the same version with nodename 10.0.0.1~n"
       "Start the new instance and run:~n"
       "scripts/restore_backup.es <<location of the backupfile>> wombat@10.0.0.1~n~n"
       "If Wombat is configured to use a different cookie other than the default 'wombat',~n"
       "the optional third parameter must be given to be used as cookie. For example:~n"
       "~s wombat@127.0.0.1 wombat@10.0.0.1 mycookie~n", [?FILE, ?FILE, ?FILE]),
    ok;
main([From, To]) ->
    main([From, To, "wombat"]);
main([From, To, CookieStr]) ->
    Node = list_to_atom(From),
    Cookie = list_to_atom(CookieStr),
    ?W("Node: ~p~n", [Node]),
    erlang:set_cookie(Node, Cookie),
    case net_kernel:connect_node(Node) of
        true ->
            ok;
        _Error ->
            ?W("WombatOAM expected to be running as: ~p~n", [Node]),
            ?W("If your instance uses a different cookie (for example: mycookie), set it by:~n"
               "~s ~s ~s mycookie~n", [?FILE, From, To]),
            halt(1)
    end,
    {ok, Pwd} = rpc:call(Node, file, get_cwd, []),
    ?W("Working directory: ~p~n", [Pwd]),
    BackupFile1 = Pwd ++ "/" ++ backup_name(From),
    BackupFile2 = Pwd ++ "/" ++ backup_name(To),
    case rpc:call(Node, mnesia, backup, [BackupFile1]) of
        ok ->
            ?W("Mnesia backup successfully created at:~n~s~n", [BackupFile1]);
        Error1 ->
            ?W("Creating a backup file failed with: ~p~n", [Error1]),
            halt(1)
    end,
    NewNode = list_to_atom(To),
    ?W("Renaming nodename from ~p to ~p ...~n", [Node, NewNode]),
    case change_node_name(mnesia_backup, Node, NewNode, BackupFile1, BackupFile2) of
        {ok, switched} ->
            ?W("Renaming of backup succeded. Updated backup is at:~n~s~n", [BackupFile2]);
        Error2 ->
            ?W("Renaming of backup failed with: ~p", [Error2]),
            halt(1)
    end,
    %% Restore db from the backup
    ?W("~nOn the new machine run:~nscripts/restore_backup.es <path_to>/~s ~s~n" , [backup_name(To), NewNode]),
    ok.

backup_name(NodeName) ->
    ?BACKUP_NAME ++ "_" ++ NodeName.

%% From http://stackoverflow.com/questions/463400/how-to-rename-the-node-running-a-mnesia-database
change_node_name(Mod, From, To, Source, Target) ->
    Switch =
        fun(Node) when Node == From -> To;
           (Node) when Node == To -> throw({error, already_exists});
           (Node) -> Node
        end,
    Convert =
        fun({schema, db_nodes, Nodes}, Acc) ->
                {[{schema, db_nodes, lists:map(Switch,Nodes)}], Acc};
           ({schema, version, Version}, Acc) ->
                {[{schema, version, Version}], Acc};
           ({schema, cookie, Cookie}, Acc) ->
                {[{schema, cookie, Cookie}], Acc};
           ({schema, Tab, CreateList}, Acc) ->
                Keys = [ram_copies, disc_copies, disc_only_copies],
                OptSwitch =
                    fun({Key, Val}) ->
                            case lists:member(Key, Keys) of
                                true -> {Key, lists:map(Switch, Val)};
                                false-> {Key, Val}
                            end
                    end,
                {[{schema, Tab, lists:map(OptSwitch, CreateList)}], Acc};
           (Other, Acc) ->
                {[Other], Acc}
        end,
    mnesia:traverse_backup(Source, Mod, Target, Mod, Convert, switched).
